from typing import Dict, Union


Metric = float
ParamsDict = Dict[str, Union[float, Metric]]  

def estimate_finetuning_performance(
    scaling_factor: str,
    model_size: int = None,
    dataset_size: int = None,
    lora_parameters: int = None,
    finetuning_dataset_size: int = None
) -> ParamsDict:
    """Estimate finetuning loss based on scaling factor and other parameters
    
    Args:
        scaling_factor: 'Model Size', 'Dataset Size' or 'LoRa Parameters'
        model_size: Model parameter size in billions, required if scaling by model size
        dataset_size: Dataset token count, required if scaling by model size 
        lora_parameters: LoRA parameter count, required if scaling by lora parameters  
        finetuning_dataset_size: Finetuning dataset token count
        
    Returns: 
        Dict with finetuning loss and scaling parameters
    """
    
    params = _get_default_params()
    params['selected_scaling_factor'] = scaling_factor

    if scaling_factor == 'Model Size':
        if model_size is None or dataset_size is None:
            raise ValueError("Please provide model size and dataset size for this scaling factor")
        params = _scale_by_model_size(params, model_size, dataset_size)

    elif scaling_factor == 'Dataset Size':   
        if dataset_size is None or finetuning_dataset_size is None:
            raise ValueError("Please provide dataset size and finetuning dataset size for this scaling factor")
        params = _scale_by_dataset_size(params, dataset_size, finetuning_dataset_size)
        
    elif scaling_factor == 'LoRa Parameters':
        if lora_parameters is None or finetuning_dataset_size is None:
            raise ValueError("Please provide LoRA parameters and finetuning dataset size for this scaling factor")
        params = _scale_by_lora_params(params, lora_parameters, finetuning_dataset_size)
        
    return params
    

def _get_default_params() -> ParamsDict:
    return {
        'A_m': 2.1*10**3,  
        'A_p': 1.4 * 10 ** 2,
        'A_t': 1.4,
        'E': 0.62,
        'alpha_m': 0.36,
        'alpha_p': 0.18, 
        'alpha_t': -0.0017,
        'beta': 0.081,
        'L': 0.0,
        'A': 0.0,
        'alpha': 0.0
    }

def _scale_by_model_size(params: ParamsDict, X: int, D: int) -> ParamsDict:
    A = params['A_m']  
    alpha = params['alpha_m']
    E = params['E']
    beta = params['beta']
    
    L = A * (1/X**alpha) * (1/D**beta) + E
    
    params['X'] = X 
    params['D'] = D
    params['L'] = L
    params['A'] = A
    params['alpha'] = alpha
    
    return params

def _scale_by_dataset_size(params: ParamsDict, X: int, D: int) -> ParamsDict:
    A = params['A_p']
    alpha = params['alpha_p']
    E = params['E']
    beta = params['beta']
    
    L = A * (1/X**alpha) * (1/D**beta) + E
    
    params['L'] = L
    params['A'] = A
    params['alpha'] = alpha
    
    return params

def _scale_by_lora_params(params: ParamsDict, X: int, D: int) -> ParamsDict:
    A = params['A_t']
    alpha = params['alpha_t']
    E = params['E']
    beta = params['beta']
    
    L = A * (1/X**alpha) * (1/D**beta) + E
    
    params['L'] = L
    params['A'] = A
    params['alpha'] = alpha
    
    return params



# Other scaling functions    

def main():
    params = estimate_finetuning_performance(
        scaling_factor='Model Size',
        model_size=4,
        dataset_size=1e6
    )
    print(f'The Expected finetuning loss is {params["L"]}')
    return params

if __name__ == '__main__':
    main()


# cli.py
import argparse
from finetuneLora import main_finetune
import sys
print(sys.path)

def main():
    parser = argparse.ArgumentParser(description="Fine-tuning script with CLI for LLM using LoRa")
    parser.add_argument("--main_path", required=True, help="Main path where the experiment will be conducted")
    parser.add_argument("--experiment_name", required=True, help="Name of the experiment")
    parser.add_argument("--training_data_path", required=True, help="Path to the training data CSV file")
    args = parser.parse_args()
    
    main_finetune(args.main_path, args.experiment_name, args.training_data_path)

if __name__ == "__main__":
    main()





import os
import torch
import argparse
from peft import set_peft_model_state_dict
from transformers import AutoConfig
from transformers.trainer_utils import PREFIX_CHECKPOINT_DIR
from transformers import TrainerCallback, TrainingArguments, TrainerState, TrainerControl

class SavePeftModelCallback(TrainerCallback):
    def on_save(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        checkpoint_folder = os.path.join(args.output_dir, f"{PREFIX_CHECKPOINT_DIR}-{state.global_step}")

        kwargs["model"].save_pretrained(checkpoint_folder)

        pytorch_model_path = os.path.join(checkpoint_folder, "pytorch_model.bin")
        torch.save({}, pytorch_model_path)
        return control


class LoadBestPeftModelCallback(TrainerCallback):
    def on_train_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        print(f"Loading best peft model from {state.best_model_checkpoint} (score: {state.best_metric}).")
        best_model_path = os.path.join(state.best_model_checkpoint, "adapter_model.bin")
        adapters_weights = torch.load(best_model_path)
        model = kwargs["model"]
        set_peft_model_state_dict(model, adapters_weights)
        return control
    


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model_path", type=str, default="Salesforce/codegen-350M-mono")
    # parser.add_argument("--model_path", type=str, default="ammarnasr/codegen-350M-mono_the-stack-dedup_java_train_peft")
    parser.add_argument("--size_valid_set", type=int, default=100)
    parser.add_argument("--seq_length", type=int, default=512)
    parser.add_argument("--max_steps", type=int, default=10000)
    parser.add_argument("--output_dir", type=str, default="./checkpoints_full")
    parser.add_argument("--log_freq", default=5, type=int)
    parser.add_argument("--eval_freq", default=1000, type=int)
    parser.add_argument("--save_freq", default=1000, type=int)
    parser.add_argument("--ft_type", type=str, default="full")










    # parser.add_argument("--model_path", type=str, default="")
    # parser.add_argument("--model_path", type=str, default="bigcode/starcoder")
    # parser.add_argument("--model_path", type=str, default="bigcode/large-model")
    # parser.add_argument("--model_path", type=str, default="codeparrot/codeparrot-small")
    # parser.add_argument("--model_path", type=str, default="bigcode/tiny_starcoder_py")
    
    # parser.add_argument("--dataset_name", type=str, default="HuggingFaceH4/CodeAlpaca_20K")
    # parser.add_argument("--subset", type=str, default=None)
    # parser.add_argument("--input_column_name", type=str, default="prompt")

    parser.add_argument("--dataset_name", type=str, default="bigcode/the-stack-dedup")
    parser.add_argument("--subset", type=str, default="data/java")
    parser.add_argument("--input_column_name", type=str, default="content")
    
    parser.add_argument("--split", type=str, default="train")
    parser.add_argument("--streaming", action="store_true", default=True)
    parser.add_argument("--shuffle_buffer", type=int, default=5000)

    # parser.add_argument("--ft_type", type=str, default="full")


    parser.add_argument("--output_column_name", type=str, default="completion")
    parser.add_argument("--batch_size", type=int, default=1)
    parser.add_argument("--gradient_accumulation_steps", type=int, default=16)
    parser.add_argument("--eos_token_id", type=int, default=49152)

    parser.add_argument("--lora_r", type=int, default=16)
    parser.add_argument("--lora_alpha", type=int, default=32)
    parser.add_argument("--lora_dropout", type=float, default=0.05)

    parser.add_argument("--learning_rate", type=float, default=5e-6)
    parser.add_argument("--lr_scheduler_type", type=str, default="cosine")
    parser.add_argument("--num_warmup_steps", type=int, default=100)
    parser.add_argument("--weight_decay", type=float, default=0.05)

    parser.add_argument("--local_rank", type=int, default=0)
    parser.add_argument("--no_fp16", action="store_false", default=True)
    parser.add_argument("--bf16", action="store_true", default=True)
    parser.add_argument("--no_gradient_checkpointing", action="store_false", default=False)
    parser.add_argument("--seed", type=int, default=0)
    parser.add_argument("--num_workers", type=int, default=None)

    

    return parser.parse_args()




import os
import json
import torch
import pickle
import jsonlines
from tqdm.auto import tqdm
from torch.utils.data import IterableDataset
from datasets import load_dataset




def prepare_sample_text(example, input_column_name="prompt", output_column_name="completion"):
    """Prepare the text from a sample of the dataset."""
    if input_column_name == "content":
        text = example[input_column_name]
        text = text[:2000]
    else:
        text = f"Question: {example[input_column_name]}\n\nAnswer: {example[output_column_name]}"
    return text




def chars_token_ratio(dataset, tokenizer, input_column_name="prompt", output_column_name="completion", nb_examples=400):
    """
    Estimate the average number of characters per token in the dataset.
    """
    total_characters, total_tokens = 0, 0
    for _, example in tqdm(zip(range(nb_examples), iter(dataset)), total=nb_examples):
        text = prepare_sample_text(example, input_column_name, output_column_name)
        total_characters += len(text)
        if tokenizer.is_fast:
            total_tokens += len(tokenizer(text).tokens())
        else:
            total_tokens += len(tokenizer.tokenize(text))

    return total_characters / total_tokens


class ConstantLengthDataset(IterableDataset):
    """
    Iterable dataset that returns constant length chunks of tokens from stream of text files.
        Args:
            tokenizer (Tokenizer): The processor used for proccessing the data.
            dataset (dataset.Dataset): Dataset with text files.
            infinite (bool): If True the iterator is reset after dataset reaches end else stops.
            seq_length (int): Length of token sequences to return.
            num_of_sequences (int): Number of token sequences to keep in buffer.
            chars_per_token (int): Number of characters per token used to estimate number of tokens in text buffer.
    """

    def __init__(
        self,
        tokenizer,
        dataset,
        infinite=False,
        seq_length=1024,
        num_of_sequences=1024,
        chars_per_token=3.6,
        input_column_name="prompt",
        output_column_name="completion"
    ):
        self.tokenizer = tokenizer
        # self.concat_token_id = tokenizer.eos_token_id if tokenizer.eos_token_id is not None else args.eos_token_id
        self.concat_token_id = tokenizer.eos_token_id
        self.dataset = dataset
        self.seq_length = seq_length
        self.infinite = infinite
        self.current_size = 0
        self.max_buffer_size = seq_length * chars_per_token * num_of_sequences
        self.input_column_name = input_column_name
        self.output_column_name = output_column_name

    def __iter__(self):
        iterator = iter(self.dataset)
        more_examples = True
        while more_examples:
            buffer, buffer_len = [], 0
            while True:
                if buffer_len >= self.max_buffer_size:
                    break
                try:
                    buffer.append(prepare_sample_text(next(iterator), self.input_column_name, self.output_column_name))
                    buffer_len += len(buffer[-1])
                except StopIteration:
                    if self.infinite:
                        iterator = iter(self.dataset)
                    else:
                        more_examples = False
                        break
            tokenized_inputs = self.tokenizer(buffer, truncation=False)["input_ids"]
            all_token_ids = []
            for tokenized_input in tokenized_inputs:
                all_token_ids.extend(tokenized_input + [self.concat_token_id])
            for i in range(0, len(all_token_ids), self.seq_length):
                input_ids = all_token_ids[i : i + self.seq_length]
                if len(input_ids) == self.seq_length:
                    self.current_size += 1
                    yield {
                        "input_ids": torch.LongTensor(input_ids),
                        "labels": torch.LongTensor(input_ids),
                    }


def create_datasets(tokenizer, args):
    dataset = load_dataset(
        args.dataset_name,
        data_dir=args.subset,
        split=args.split,
        use_auth_token=True,
        num_proc=args.num_workers if not args.streaming else None,
        streaming=args.streaming,
    )
    if args.streaming:
        print("Loading the dataset in streaming mode")
        valid_data = dataset.take(args.size_valid_set)
        train_data = dataset.skip(args.size_valid_set)
        train_data = train_data.shuffle(buffer_size=args.shuffle_buffer, seed=args.seed)
    else:
        train_data = dataset["train"]
        valid_data = dataset["test"]
        print(f"Size of the train set: {len(train_data)}. Size of the validation set: {len(valid_data)}")

    chars_per_token = chars_token_ratio(train_data, tokenizer, args.input_column_name, args.output_column_name)
    print(f"The character to token ratio of the dataset is: {chars_per_token:.2f}")

    train_dataset = ConstantLengthDataset(
        tokenizer,
        train_data,
        infinite=True,
        seq_length=args.seq_length,
        chars_per_token=chars_per_token,
        input_column_name=args.input_column_name,
        output_column_name=args.output_column_name
    )
    valid_dataset = ConstantLengthDataset(
        tokenizer,
        valid_data,
        infinite=False,
        seq_length=args.seq_length,
        chars_per_token=chars_per_token,
        input_column_name=args.input_column_name,
        output_column_name=args.output_column_name
    )
    return train_dataset, valid_dataset



def convert_json_to_jsonlines(filename):
    with open(filename) as f:
        data = json.load(f)
    filename_without_ext = filename[:-5]
    jsonl_filename = filename_without_ext+'.jsonl'
    with jsonlines.open(jsonl_filename, mode='w') as writer:
        for i in tqdm(range(len(data['hexsha']))):
            writer.write({
                'hexsha': data['hexsha'][i],
                'size': data['size'][i],
                'content': data['content'][i],
                'avg_line_length': data['avg_line_length'][i],
                'max_line_length': data['max_line_length'][i],
                'alphanum_fraction': data['alphanum_fraction'][i]
            })
    return jsonl_filename

def push_to_huggingface(num_samples=1000000):
    print('Loading dataset from pickle file')
    trainfile = './data_pkl/java/bigcode-the-stack-dedup-train.pkl'
    with open(trainfile, "rb") as f:
        train_ds = pickle.load(f)
    
    train_ds_dict = {
        'hexsha': [],
        'size': [],
        'content': [],
        'avg_line_length': [],
        'max_line_length': [],
        'alphanum_fraction': []
    }
    save_every = num_samples//10
    trainfile_json = f'./data_pkl/java/bigcode-the-stack-dedup-train.json'
    length = len(train_ds['hexsha'])

    print(f'Creating json file with {num_samples} samples which is {num_samples/length*100}% of the dataset')
    current_size = 0

    tbar = tqdm(range(num_samples),total=num_samples, desc=f'latest file size: {current_size}MB')

    for i in tbar:
        train_ds_dict['hexsha'].append(train_ds[i]['hexsha'])
        train_ds_dict['size'].append(train_ds[i]['size'])
        train_ds_dict['content'].append(train_ds[i]['content'])
        train_ds_dict['avg_line_length'].append(train_ds[i]['avg_line_length'])
        train_ds_dict['max_line_length'].append(train_ds[i]['max_line_length'])
        train_ds_dict['alphanum_fraction'].append(train_ds[i]['alphanum_fraction'])
        if i % save_every == 0:
            with open(trainfile_json, 'w') as f:
                json.dump(train_ds_dict, f)
            current_size = os.path.getsize(trainfile_json)/1e6
            tbar.set_description(f'latest file size: {current_size:.2f}MB')
            

    print('Converting json to jsonl')
    trainfile_jsonl = convert_json_to_jsonlines(trainfile_json)

    print('Pushing to huggingface')
    hf_ds = load_dataset("json", data_files=trainfile_jsonl)
    hf_repo = 'ammarnasr/bigcode-the-stack-dedup-java-small-subset'
    hf_ds.push_to_hub(hf_repo)



class ConstantLengthDataset(IterableDataset):
    """
    Iterable dataset that returns constant length chunks of tokens from stream of text files.
        Args:
            tokenizer (Tokenizer): The processor used for proccessing the data.
            dataset (dataset.Dataset): Dataset with text files.
            infinite (bool): If True the iterator is reset after dataset reaches end else stops.
            seq_length (int): Length of token sequences to return.
            num_of_sequences (int): Number of token sequences to keep in buffer.
            chars_per_token (int): Number of characters per token used to estimate number of tokens in text buffer.
    """

    def __init__(
        self,
        tokenizer,
        dataset,
        infinite=False,
        seq_length=1024,
        num_of_sequences=1024,
        chars_per_token=2.95,
    ):
        self.tokenizer = tokenizer
        self.concat_token_id = tokenizer.eos_token_id if tokenizer.eos_token_id else 49152
        self.dataset = dataset
        self.seq_length = seq_length
        self.infinite = infinite
        self.current_size = 0
        self.max_buffer_size = seq_length * chars_per_token * num_of_sequences
        self.content_field = "content"

    def __iter__(self):
        iterator = iter(self.dataset)
        more_examples = True
        while more_examples:
            buffer, buffer_len = [], 0
            while True:
                if buffer_len >= self.max_buffer_size:
                    break
                try:
                    buffer.append(next(iterator)[self.content_field])
                    buffer_len += len(buffer[-1])
                except StopIteration:
                    if self.infinite:
                        iterator = iter(self.dataset)
                    else:
                        more_examples = False
                        break
            tokenized_inputs = self.tokenizer(buffer, truncation=False)["input_ids"]
            all_token_ids = []
            for tokenized_input in tokenized_inputs:
                all_token_ids.extend(tokenized_input + [self.concat_token_id])
            for i in range(0, len(all_token_ids), self.seq_length):
                input_ids = all_token_ids[i : i + self.seq_length]
                if len(input_ids) == self.seq_length:
                    self.current_size += 1
                    yield {
                        "input_ids": torch.LongTensor(input_ids),
                        "labels": torch.LongTensor(input_ids),
                    }






import dataclasses
from typing import Dict, List

@dataclasses.dataclass
class GPUTensorCoreSpec:
    """Specifications of a GPU tensor core."""
    name: str
    clock_rate_ghz: float
    num_tensor_cores: int
    flops_per_cycle_fp32: int = 64
    flops_per_cycle_fp16: int = 128

@dataclasses.dataclass  
class GPUSpec:
    """Specifications of a GPU."""
    name: str
    architecture: str
    gpus: List[GPUTensorCoreSpec]

@dataclasses.dataclass
class HardwareSpec:
    """Specifications of available hardware."""
    gpus: List[GPUSpec]


def calculate_gpu_throughput(gpu: GPUTensorCoreSpec, 
                             precision: str = "fp32") -> float:
    """Calculate throughput of a GPU in TFLOPS."""
    flops_per_cycle = (gpu.flops_per_cycle_fp32 
                       if precision == "fp32" 
                       else gpu.flops_per_cycle_fp16)  
    throughput = (gpu.clock_rate_ghz * 1e9 * gpu.num_tensor_cores * 
                  flops_per_cycle)
    return throughput / 1e12


def calculate_aggregate_throughput(hardware: HardwareSpec,
                                   precision: str = "fp32") -> float:
    """Calculate aggregated throughput across GPUs in TFLOPS.""" 
    throughput = 0
    for gpu_spec in hardware.gpus:
        for gpu in gpu_spec.gpus:
            throughput += calculate_gpu_throughput(gpu, precision)
    return throughput


@dataclasses.dataclass   
class ExperimentSpec:
    """Specifications of a training experiment."""
    model_params: int  
    dataset: int
    hardware: HardwareSpec
    precision: str = "fp32"
    hours_trained: float = 1.0 # training hours
    

def calculate_compute_cost(spec: ExperimentSpec) -> float:
    """Calculate expected compute cost of an experiment."""
    
    throughput = calculate_aggregate_throughput(spec.hardware, 
                                                spec.precision)
                                                
    expected_compute = 6 * spec.model_params * spec.dataset  
    actual_compute = throughput * 1e12 * spec.hours_trained  
    
    return {
        "Required FLOPs": expected_compute / 1e12,
        "Actual FLOPs": actual_compute / 1e12,
    }

def main():
    # Define GPU specifications
    gpus = [
        GPUTensorCoreSpec(name="A100", clock_rate_ghz=1.41, 
                          num_tensor_cores=6912),
        GPUTensorCoreSpec(name="V100", clock_rate_ghz=1.53, 
                          num_tensor_cores=5120),
    ]
    gpu_specs = GPUSpec(name="NVIDIA", architecture="Ampere", gpus=gpus)
    
    # Define hardware specifications
    hardware = HardwareSpec(gpus=[gpu_specs])
    
    # Define experiment specifications
    spec = ExperimentSpec(model_params=1e9, dataset=1e12, 
                           hardware=hardware, precision="fp32", 
                           hours_trained=1.0)
    
    # Calculate compute cost
    cost = calculate_compute_cost(spec)
    print(f"Compute cost: {cost:.2f} TFLOPS")


















# import torch
# import pandas as pd
# import plotly.graph_objects as go
# import sys
# import ctypes

# # Some constants taken from cuda.h
# CUDA_SUCCESS = 0
# CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT = 16
# CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR = 39
# CU_DEVICE_ATTRIBUTE_CLOCK_RATE = 13
# CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE = 36

# # GPU Database containing the information about the most common GPUs.
# GPU_DATABASE ={
#     "NVIDIA GeForce GTX 1650": {
#         "Name": "NVIDIA GeForce GTX 1650",
#         "Compute Capability": "7.5",
#         "Multiprocessors": 14,
#         "CUDA Cores": 896,
#         "CUDA Architecture": "Turing",
#         "Ops per cycle": {
#             "FP64": 2,
#             "FP32": 64,
#             "FP16": 128,
#             "INT8": 256
#         },
#         "Concurrent threads": 14336,
#         "GPU clock": 1515.0,
#         "Memory clock": 6001.0,
#         "Total Memory (MiB)": 4095.6875,
#     },
#     "NVIDIA T4":{
#         "Name": "NVIDIA T4",
#         "Compute Capability": "7.5",
#         "Multiprocessors": 40,
#         "CUDA Cores": 2560,
#         "CUDA Architecture": "Turing",
#         "Ops per cycle": {
#             "FP64": 2,
#             "FP32": 64,
#             "FP16": 128,
#             "INT8": 256
#         },
#         "Concurrent threads": 40960,
#         "GPU clock": 1590.0,
#         "Memory clock": 5001.0,
#         "Total Memory (MiB)": 15102.0625,
#     },
#     "NVIDIA Tesla P100-PCIE-16GB":{
#         "Name": "NVIDIA Tesla P100-PCIE-16GB",
#         "Compute Capability": "6.0",
#         "Multiprocessors": 56,
#         "CUDA Cores": 3584,
#         "CUDA Architecture": "Pascal",
#         "Ops per cycle": {
#             "FP64": 32,
#             "FP32": 128,
#             "FP16": 256,
#             "INT8": 512
#         },
#         "Concurrent threads": 100352,
#         "GPU clock": 1328.0,
#         "Memory clock": 715.0,
#         "Total Memory (MiB)": 16276.0,
#     },
#     "NVIDIA Tesla V100-SXM2-16GB":{
#         "Name": "NVIDIA Tesla V100-SXM2-16GB",
#         "Compute Capability": "7.0",
#         "Multiprocessors": 80,
#         "CUDA Cores": 5120,
#         "CUDA Architecture": "Volta",
#         "Ops per cycle": {
#             "FP64": 2,
#             "FP32": 64,
#             "FP16": 128,
#             "INT8": 256
#         },
#         "Concurrent threads": 409600,
#         "GPU clock": 1530.0,
#         "Memory clock": 877.0,
#         "Total Memory (MiB)": 16130.0,
#     },
#     "NVIDIA A100-SXM4-40GB":{
#         "Name": "NVIDIA A100-SXM4-40GB",
#         "Compute Capability": "8.0",
#         "Multiprocessors": 108,
#         "CUDA Cores": 6912,
#         "CUDA Architecture": "Ampere",
#         "Ops per cycle": {
#             "FP64": 32,
#             "FP32": 128,
#             "FP16": 256,
#             "INT8": 512
#         },
#         "Concurrent threads": 746496,
#         "GPU clock": 1410.0,
#         "Memory clock": 1556.0,
#         "Total Memory (MiB)": 40536.0,
#     },
# }

# def list_common_gpus():
#     return list(GPU_DATABASE.keys())

# def ConvertSMVer2ArchName(major, minor):
#     # Returns the name of a CUDA architecture given a device capability
#     # (major, minor) pair.
#     # See _ConvertSMVer2ArchName in helper_cuda.h in NVIDIA's CUDA Samples.
#     return {(1, 0): "Tesla",
#             (1, 1): "Tesla",
#             (1, 2): "Tesla",
#             (1, 3): "Tesla",
#             (2, 0): "Fermi",
#             (2, 1): "Fermi",
#             (3, 0): "Kepler",
#             (3, 2): "Kepler",
#             (3, 5): "Kepler",
#             (3, 7): "Kepler",
#             (5, 0): "Maxwell",
#             (5, 2): "Maxwell",
#             (5, 3): "Maxwell",
#             (6, 0): "Pascal",
#             (6, 1): "Pascal",
#             (6, 2): "Pascal",
#             (7, 0): "Volta",
#             (7, 2): "Volta",
#             (7, 5): "Turing",
#             (8, 0): "Ampere",
#             (8, 6): "Ampere",
#             (8, 7): "Ampere",
#             (8, 9): "Ampere",
#             (9, 0): "Hopper",
#             }.get((major, minor), "Unknown")

# def ConvertSMVer2Cores(major, minor):
#     # Returns the number of CUDA cores per multiprocessor for a given
#     # Compute Capability version. There is no way to retrieve that via
#     # the API, so it needs to be hard-coded.
#     # See _ConvertSMVer2Cores in helper_cuda.h in NVIDIA's CUDA Samples.
#     return {(1, 0): 8,    # Tesla
#             (1, 1): 8,
#             (1, 2): 8,
#             (1, 3): 8,
#             (2, 0): 32,   # Fermi
#             (2, 1): 48,
#             (3, 0): 192,  # Kepler
#             (3, 2): 192,
#             (3, 5): 192,
#             (3, 7): 192,
#             (5, 0): 128,  # Maxwell
#             (5, 2): 128,
#             (5, 3): 128,
#             (6, 0): 64,   # Pascal
#             (6, 1): 128,
#             (6, 2): 128,
#             (7, 0): 64,   # Volta
#             (7, 2): 64,
#             (7, 5): 64,   # Turing
#             (8, 0): 64,   # Ampere
#             (8, 6): 128,
#             (8, 7): 128,
#             (8, 9): 128,  # Ada
#             (9, 0): 128,  # Hopper
#             }.get((major, minor), 0)

# def ConvertArchName2OpsPerSM(arch_name, precision):
#     return {('Volta', 'FP64'): 2,
#             ('Volta', 'FP32'): 64,
#             ('Volta', 'FP16'): 128,
#             ('Volta', 'INT8'): 256,
#             ('Turing', 'FP64'): 2,
#             ('Turing', 'FP32'): 64,
#             ('Turing', 'FP16'): 128,
#             ('Turing', 'INT8'): 256,
#             ('Ampere', 'FP64'): 32,
#             ('Ampere', 'FP32'): 128,
#             ('Ampere', 'FP16'): 256,
#             ('Ampere', 'INT8'): 512,
#     }.get((arch_name, precision), 0)

# def cuda_main(verbose=True):
#     info_dict = {}
#     libnames = ('libcuda.so', 'libcuda.dylib', 'nvcuda.dll', 'cuda.dll')
#     for libname in libnames:
#         try:
#             cuda = ctypes.CDLL(libname)
#         except OSError:
#             continue
#         else:
#             break
#     else:
#         raise OSError("could not load any of: " + ' '.join(libnames))
#     nGpus = ctypes.c_int()
#     name = b' ' * 100
#     cc_major = ctypes.c_int()
#     cc_minor = ctypes.c_int()
#     cores = ctypes.c_int()
#     threads_per_core = ctypes.c_int()
#     clockrate = ctypes.c_int()
#     freeMem = ctypes.c_size_t()
#     totalMem = ctypes.c_size_t()

#     result = ctypes.c_int()
#     device = ctypes.c_int()
#     context = ctypes.c_void_p()
#     error_str = ctypes.c_char_p()

#     result = cuda.cuInit(0)
#     if result != CUDA_SUCCESS:
#         cuda.cuGetErrorString(result, ctypes.byref(error_str))
#         print("cuInit failed with error code %d: %s" % (result, error_str.value.decode()))
#         return 1
#     result = cuda.cuDeviceGetCount(ctypes.byref(nGpus))
#     if result != CUDA_SUCCESS:
#         cuda.cuGetErrorString(result, ctypes.byref(error_str))
#         print("cuDeviceGetCount failed with error code %d: %s" % (result, error_str.value.decode()))
#         return 1
#     if verbose:
#         print("Found %d device(s)." % nGpus.value)
#     for i in range(nGpus.value):
#         result = cuda.cuDeviceGet(ctypes.byref(device), i)
#         if result != CUDA_SUCCESS:
#             cuda.cuGetErrorString(result, ctypes.byref(error_str))
#             print("cuDeviceGet failed with error code %d: %s" % (result, error_str.value.decode()))
#             return 1
#         if verbose:
#             print("Device: %d" % i)
#         info_dict[f'device_{i}'] = {}
#         if cuda.cuDeviceGetName(ctypes.c_char_p(name), len(name), device) == CUDA_SUCCESS:
#             device_name = name.split(b'\0', 1)[0].decode()
#             if verbose:
#                 print("  Name: %s" % device_name)
#             info_dict[f'device_{i}']['Name'] = device_name
#         if cuda.cuDeviceComputeCapability(ctypes.byref(cc_major), ctypes.byref(cc_minor), device) == CUDA_SUCCESS:
#             compute_capability = f'{cc_major.value}.{cc_minor.value}'
#             if verbose:
#                 print("  Compute Capability: %s" % compute_capability)
#             info_dict[f'device_{i}']['Compute Capability'] = compute_capability
#         if cuda.cuDeviceGetAttribute(ctypes.byref(cores), CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT, device) == CUDA_SUCCESS:
#             multiprocessors = cores.value
#             cuda_cores = cores.value * ConvertSMVer2Cores(cc_major.value, cc_minor.value)
#             cuda_arch = ConvertSMVer2ArchName(cc_major.value, cc_minor.value)
#             if verbose:
#                 print("  Multiprocessors: %d" % multiprocessors)
#                 print("  CUDA Cores: %s" % (cuda_cores or "unknown"))
#                 print("  CUDA Architecture: %s" % cuda_arch)
#             info_dict[f'device_{i}']['Multiprocessors'] = multiprocessors
#             info_dict[f'device_{i}']['CUDA Cores'] = cuda_cores
#             info_dict[f'device_{i}']['CUDA Architecture'] = cuda_arch

#             presicions = ['FP64', 'FP32', 'FP16', 'INT8']
#             info_dict[f'device_{i}']['Ops per cycle'] = {}
#             for precision in presicions:
#                 ops_per_cycle = ConvertArchName2OpsPerSM(cuda_arch, precision)
#                 if verbose:
#                     print(f"  {precision} ops per cycle: {ops_per_cycle}")
#                 info_dict[f'device_{i}']['Ops per cycle'][precision] = ops_per_cycle

                

#             if cuda.cuDeviceGetAttribute(ctypes.byref(threads_per_core), CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR, device) == CUDA_SUCCESS:
#                 concurrrent_threads = cores.value * threads_per_core.value
#                 if verbose:
#                     print("  Concurrent threads: %d" % concurrrent_threads)
#                 info_dict[f'device_{i}']['Concurrent threads'] = concurrrent_threads
#         if cuda.cuDeviceGetAttribute(ctypes.byref(clockrate), CU_DEVICE_ATTRIBUTE_CLOCK_RATE, device) == CUDA_SUCCESS:
#             gpu_clockrate = clockrate.value / 1000.
#             if verbose:
#                 print("  GPU clock: %g MHz" % gpu_clockrate)
#             info_dict[f'device_{i}']['GPU clock'] = gpu_clockrate
#         if cuda.cuDeviceGetAttribute(ctypes.byref(clockrate), CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE, device) == CUDA_SUCCESS:
#             memory_clockrate = clockrate.value / 1000.
#             if verbose:
#                 print("  Memory clock: %g MHz" % memory_clockrate)
#             info_dict[f'device_{i}']['Memory clock'] = memory_clockrate
#         try:
#             result = cuda.cuCtxCreate_v2(ctypes.byref(context), 0, device)
#         except AttributeError:
#             result = cuda.cuCtxCreate(ctypes.byref(context), 0, device)
#         if result != CUDA_SUCCESS:
#             cuda.cuGetErrorString(result, ctypes.byref(error_str))
#             print("cuCtxCreate failed with error code %d: %s" % (result, error_str.value.decode()))
#         else:
#             try:
#                 result = cuda.cuMemGetInfo_v2(ctypes.byref(freeMem), ctypes.byref(totalMem))
#             except AttributeError:
#                 result = cuda.cuMemGetInfo(ctypes.byref(freeMem), ctypes.byref(totalMem))
#             if result == CUDA_SUCCESS:
#                 total_memory = totalMem.value / 1024**2
#                 free_memory = freeMem.value / 1024**2
#                 if verbose:
#                     print("  Total Memory: %ld MiB" % total_memory)
#                     print("  Free Memory: %ld MiB" % free_memory)
#                 info_dict[f'device_{i}']['Total Memory (MiB)'] = total_memory
#                 # info_dict[f'device_{i}']['Free Memory (MiB)'] = free_memory
#             else:
#                 cuda.cuGetErrorString(result, ctypes.byref(error_str))
#                 print("cuMemGetInfo failed with error code %d: %s" % (result, error_str.value.decode()))
#             cuda.cuCtxDetach(context)
#     return info_dict


# def gpu_flops(gpu_info, precision=1):
#     precision_map = {
#         1: 'FP64',
#         2: 'FP32',
#         4: 'FP16',
#     }
#     clock_rate = gpu_info['GPU clock'] # in MHz
#     clock_rate *= 1000**2 # in Hz
#     cores = gpu_info['CUDA Cores']
#     float_precision = precision_map[precision]
#     flops_per_clock_cycle = gpu_info['Ops per cycle'][float_precision]
#     flops = clock_rate * cores * flops_per_clock_cycle * precision
#     tera_flops = flops/1000**4
#     return tera_flops

# def custom_gpu_info(gpu_name):
#     if gpu_name in GPU_DATABASE:
#         return GPU_DATABASE[gpu_name]
#     else:
#         print(f'GPU {gpu_name} not found in database')
#         return None
    
# def print_large_number(number, bytebased=False):
#     '''
#     Utility function to return a string representation of a large number
#     It uses scales (Trillion, Billion, Million, Thousand)
#     If bytebased is True, it uses scales (Terabyte, Gigabyte, Megabyte, Kilobyte)
#     '''
#     if bytebased:
#         scales = ['TB', 'GB', 'MB', 'KB']
#     else:
#         scales = ['T', 'B', 'M', 'K']
#     scale = 0
#     while number >= 1000:
#         number /= 1000
#         scale += 1
#         if scale == 3:
#             break
#     return f'{number:.2f}{scales[scale]}'


# def local_gpu_info():
#     if torch.cuda.is_available():
#         local_cuda_info = cuda_main(verbose=False)
#         num_gpus = len(local_cuda_info)
#         print(f'GPU is available! You have {num_gpus} GPUs Locally.')
#     else:
#         print('No GPU available Locally!')
#         local_cuda_info = None
#         num_gpus = 0
#     return local_cuda_info, num_gpus




# def number_of_tokens_estimator(dataset_size):
#     tokens_per_gb = 90 * 10**6
#     if 'TB' in dataset_size:
#         modifier = 10**12
#         val = float(dataset_size.replace('TB', ''))
#     elif 'GB' in dataset_size:
#         modifier = 10**9
#         val = float(dataset_size.replace('GB', ''))
#     elif 'MB' in dataset_size:
#         modifier = 10**6
#         val = float(dataset_size.replace('MB', ''))
#     elif 'KB' in dataset_size:
#         modifier = 10**3
#         val = float(dataset_size.replace('KB', ''))
#     else:
#         modifier = 1
#         val = float(dataset_size)
#     return int(val * modifier * tokens_per_gb / 10**9)
    




# def Calculate_throughput(gpu_info, float_precision='FP32'):
#     precision_map = {
#         'FP64': 1,
#         'FP32': 2,
#         'FP16': 4,
#     }
#     clock_rate = gpu_info['GPU clock'] # in MHz
#     clock_rate *= 1000**2 # in Hz
#     cores = gpu_info['CUDA Cores']
#     precision = precision_map[float_precision]
#     flops_per_clock_cycle = gpu_info['Ops per cycle'][float_precision]
#     flops = clock_rate * cores * flops_per_clock_cycle * precision
#     tera_flops = flops/1000**4
#     return tera_flops


# def calculate_tao_for_gpu_precision(devices_info, precision='FP32'):
#     tao = 0
#     for device_info in devices_info:
#         tao += Calculate_throughput(device_info, precision=precision)
#     return tao



# def calculate_dataset_size(k=0):
#     num_tokens = number_of_tokens_estimator(dataset_size)
#     return num_tokens


# def calculate_training_time(k=0):
#     #Let the user enter the training time manually
#     training_time = st.number_input(
#         'Enter the training time in hours:',
#         value=1,
#         key=k,
#         step=10,
#     )
#     k += 1
#     #Display the training time
#     st.markdown(f"##### Training Time = :violet[{training_time} hours]")

#     #Add training time to the session state
#     st.session_state['training_time'] = training_time

#     return training_time*3600



# def main(num_parameters, training_time, dataset_size, gpu_info=None, local_cuda=False, precision='FP32', scaling_factor='Model Size'):
#     P = num_parameters
#     T = training_time
#     D = number_of_tokens_estimator(dataset_size)
#     if local_cuda:
#         local_cuda_info, num_gpus = local_gpu_info()
#         if num_gpus > 0:
#             tao = calculate_tao_for_gpu_precision(local_cuda_info, precision=precision)
#         else:
#             print('No GPU available Locally, try to use gpu_info parameter to pass the GPU info')
#             tao = 0
#     else:
#         tao = calculate_tao_for_gpu_precision(gpu_info, precision=precision)


#     if D >= 20*P:
#         print(f" The dataset size is sufficient for the model parameters (D >= 20P).")
#     else:
#         print(f" The dataset size is not sufficient for the model parameters (D < 20P).")

#     print("Now we can calculate the cost using the scaling law equation:")
#     print("C = tao*T = 6*P*D")
#     print(f"tao is the Aggregated Throughput = {tao} TFLOPS")
#     print(f"T is the Training Time = {T} Seconds")
#     print(f"P is the Number of Parameters = {P} Parameters")
#     print(f"D is the Number of Tokens = {D} Tokens")
#     print(f"6 is the Scaling Law Constant")
#     lhs_cost = (tao*(1000**4))*T
#     rhs_cost = 6*P*D
#     print(f"The FLOPs based on the Hardware and Training Time = {lhs_cost} TFLOPS")
#     print(f"The FLOPs based on the Model and Dataset Size = {rhs_cost} TFLOPS")

#     if lhs_cost > rhs_cost:
#         overtrain = lhs_cost - rhs_cost
#         optimal_hours = (rhs_cost/(tao*1000**4))/3600
#         optimal_throughput = (rhs_cost/T)/10**12
#         print(f"Overtraining = {overtrain} TFLOPS")
#         print(f"You can reduce either the training time or Hardware throughput to reduce the monetary cost.")
#         print(f"Optimal Throughput = {optimal_throughput} TFLOPS")
#         print(f"Optimal Training Time = {optimal_hours} hour(s)")
#     else:
#         undertrain = rhs_cost - lhs_cost
#         optimal_hours = (rhs_cost/(tao*1000**4))/3600
#         print(f"Undertraining = {undertrain} TFLOPS")
#         print(f"To train the model with the parameters, dataset size, and hardware you have selected, you need to train for at least {optimal_hours} hour(s)")
        


#     print(f"---")
#     print("<h1 style='text-align: center; color: Tomato;'>Compute Optimal LLMs Fine Tuning</h1>", unsafe_allow_html=True)
#     print("**We use the following multiplicative joint scaling law for LLM finetuning:**")
#     print("L(X, D) = A \left( \frac{1}{X^\alpha} \right) \left( \frac{1}{D_f^\beta} \right) + E")
#     print('''
#                 **Where:**

#                     - L(X,D) represents the loss function, which quantifies the difference between the predicted output and the desired output.
                
#                     - X and D are vectors of scaling factors and the fine-tuning data size, respectively. The scaling factors represent the magnitudes of the resources used during training, such as the amount of compute or the number of training examples.
                
#                     - A, E, α, and β are the parameters of the model. A represents the base loss, E represents the constant error, α and β represent the exponents of the scaling factors.
                
#                 ''')
    




import datetime

class GitHubRepoSearchQueryBuilder:
    def __init__(self):
        self.query_params = {}

    def init_from_args(self, args):
        value = args['value']

        search_in = ''
        prefix = ' in:'
        if args['name']:
            search_in += f'{prefix}name'
            prefix = ','
        if args['description']:
            search_in += f'{prefix}description'
            prefix = ','
        if args['readme']:
            search_in += f'{prefix}readme'
            prefix = ','
        if args['topics']:
            search_in += f'{prefix}topics'
            prefix = ','

        self.query_params[value] = search_in

        if 'owner' in args and 'name' in args:
            self.query_params['repo:'] = f'{args["owner"]}/{args["name"]}'
            
        if 'user' in args:
            self.query_params['user:'] = args['user']

        if 'min_size' in args and 'max_size' in args:
            self.query_params['size:'] = f'{args["min_size"]}..{args["max_size"]}'
        elif 'min_size' in args:
            self.query_params['size:'] = f'>{args["min_size"]}'
        elif 'max_size' in args:
            self.query_params['size:'] = f'<{args["max_size"]}'

        if 'min_forks' in args and 'max_forks' in args:
            self.query_params['forks:'] = f'{args["min_forks"]}..{args["max_forks"]}'
        elif 'min_forks' in args:
            self.query_params['forks:'] = f'>{args["min_forks"]}'
        elif 'max_forks' in args:
            self.query_params['forks:'] = f'<{args["max_forks"]}'

        if 'min_stars' in args and 'max_stars' in args:
            self.query_params['stars:'] = f'{args["min_stars"]}..{args["max_stars"]}'
        elif 'min_stars' in args:
            self.query_params['stars:'] = f'>{args["min_stars"]}'
        elif 'max_stars' in args:
            self.query_params['stars:'] = f'<{args["max_stars"]}'

        if 'min_created' in args and 'max_created' in args:
            self.query_params['created:'] = f'{args["min_created"]}..{args["max_created"]}'
        elif 'min_created' in args:
            self.query_params['created:'] = f'>{args["min_created"]}'
        elif 'max_created' in args:
            self.query_params['created:'] = f'<{args["max_created"]}'

        if 'license' in args:
            self.query_params['license:'] = args['license']

        if 'language' in args:
            self.query_params['language:'] = args['language']

        if 'topic' in args:
            self.query_params['topic:'] = args['topic']


    def build(self):
        full_query = ''
        for key, value in self.query_params.items():
            full_query += f'{key}{value} '
        return full_query.strip()


def get_query_args(**args):
    """
    Builds a dictionary of query arguments based on the provided keyword arguments.
    """
    query_args = {
        'name': False,
        'description': False,
        'readme': False,
        'topics': False,
        'owner': 'skip',
        'repo_name': 'skip',
        'user': 'skip',
        'min_size': 666,
        'max_size': 666,
        'min_forks': 666,
        'max_forks': 666,
        'min_stars': 666,
        'max_stars': 666,
        'min_created': datetime.date(2000, 1, 1),
        'max_created': datetime.date(2000, 1, 1),
        'language': 'skip',
        'topic': 'skip',
        'license': 'skip'
    }

    for key, value in args.items():
        if key in query_args:
            query_args[key] = value

    kes_to_delete = []
    for key in query_args:
        if query_args[key] == 'skip':
            kes_to_delete.append(key)
        if query_args[key] == 666:
            kes_to_delete.append(key)
        if query_args[key] == datetime.date(2000, 1, 1):
            kes_to_delete.append(key)
    for key in kes_to_delete:
        del query_args[key]
    return query_args

def get_extensions_dict():
    code_extensions_column_name_dict = {
        'py': 'Python',
        'ipynb': 'Python',
        'rmd': 'R',
        'r': 'R',
        'scala': 'Scala',
        'java': 'Java',
        'js': 'JavaScript',
        'go': 'Go',

        'c': 'C',
        'cpp': 'C++',
        'cs': 'C#',


        'html': 'HTML',
        'css': 'CSS',
        'php': 'PHP',

        'rb': 'Ruby',
        'pl': 'Perl',
        'jl': 'Julia',
        'kt': 'Kotlin',
        'swift': 'Swift',
        'vb': 'Visual Basic',
        'vba': 'Visual Basic',
        'vbnet': 'Visual Basic',
        'vb.net': 'Visual Basic',
        'ts': 'TypeScript',
        'tsx': 'TypeScript',
        'jsx': 'JavaScript',
        'tsx': 'TypeScript',
        'dart': 'Dart',
        'lua': 'Lua',
        'sh': 'Shell',
        'bash': 'Shell',
        'ps1': 'PowerShell',
        'psm1': 'PowerShell',
        'psd1': 'PowerShell',
        'ps1xml': 'PowerShell',
        'psc1': 'PowerShell',
        'psrc': 'PowerShell',
        'pp': 'Pascal',
        'pas': 'Pascal',
        'pl': 'Perl',
        'pm': 'Perl',
        't': 'Perl',
        'pod': 'Perl',
        

        'sql': 'SQL',
        'sh': 'Shell',
        'json': 'JSON',
        'xml': 'XML',
        'yml': 'YAML',
        'yaml': 'YAML',
        'md': 'Markdown',
        'txt': 'Text',
        'cfg': 'Config',
        'ini': 'Config',
        'conf': 'Config',
        'cfg': 'Config',
        'gitignore': 'Config',
        'gitattributes': 'Config',
        'gitmodules': 'Config',
        'gitkeep': 'Config',
        'gitconfig': 'Config',
        'git': 'Config',
    }
    return code_extensions_column_name_dict


from setuptools import setup, find_packages
import sys
print(sys.path)
setup(
    name='llm_finetuning',
    version='0.1',
    packages=find_packages(),
    install_requires=[
        'tqdm',
        'numpy',
        'wandb',
        'pandas',
        'plotly',
        'seaborn',
        'argparse',
        'datasets',
        'jsonlines',
        'accelerate',
        'matplotlib',
        'transformers',
        'torch==1.10'
    ],
    entry_points={
        'console_scripts': [
            'llm-finetune=llm_finetuning.cli:main',
        ],
    },
)






import requests
import pandas as pd
from tqdm.auto import tqdm
from ..utils import GitHubRepoSearchQueryBuilder, get_query_args



def get_repos_in_page(page_num, token, query):
    """
    Fetches a page of repositories from GitHub based on the provided query.
    Args:
        page_num (int): The page number to fetch.
        token (str): The GitHub API token for authentication.
        query (str): The search query to use.
    Returns:
        list: A list of dictionaries containing information about the fetched repositories.
    """
    headers = {
        'Accept': 'application/vnd.github+json',
        'Authorization': f'Bearer {token}',
        'X-GitHub-Api-Version': '2022-11-28',
    }

    params = {
        'q': query,
        'page': page_num
    }

    res = requests.get('https://api.github.com/search/repositories', params=params, headers=headers)
    if res.status_code == 200:
        return res.json()["items"]
    else:
        print(f"Failed to fetch page {page_num} of repositories: {res.status_code}")
        print(res.text)
        return None

def search_github_repos(query, max_results, token):
    """
    Searches GitHub for repositories based on the given query.
    Returns a dataframe containing information about the matched repositories.

    Args:
        query (str): The search query to use.
        max_results (int): The maximum number of repositories to return.
        token (str): The GitHub API token for authentication.

    Returns:
        pandas.DataFrame: A dataframe containing information about the matched repositories.
    """
    num_pages = max_results//30
    if max_results % 30 > 0:
        num_pages += 1

    outputs = []
    for page in tqdm(range(num_pages), desc="Fetching Repositories", total=num_pages):
        print(f'fetching repositories: {min(len(outputs), max_results)}')
        repos = get_repos_in_page(page, token, query)
        if repos:
            outputs.extend(repos)
    outputs = outputs[:max_results]
    repos_df = extract_repo_info(outputs)
    print(f'fetched {len(repos_df)} repositories')
    return repos_df

def extract_repo_info(repos_list):
    """
    Extracts basic information from a list of GitHub repositories.
    Args:
        repos_list (list): A list of dictionaries containing information about GitHub repositories.

    Returns:
        pandas.DataFrame: A dataframe containing basic information about the repositories.
    """
    df = pd.DataFrame(repos_list)
    df = df[["full_name", "html_url", "description", "stargazers_count", "forks_count", "updated_at", "created_at", "owner", "topics", "size", "language", "license"]]
    return df

def fetch_repos(token, max_results, filename, search_term, **args): 
    """
    Fetches repositories from GitHub based on the provided search term and query arguments.
    Saves the fetched repositories to a CSV file.

    Args:
        token (str): The GitHub API token for authentication.
        max_results (int): The maximum number of repositories to return.
        filename (str): The name of the file to save the fetched repositories to.
        search_term (str): The search term to use.
        **args: Additional keyword arguments to use as query parameters these include:
            - name (bool): if True, the search term will be matched against the repository name.
            - description (bool): if True, the search term will be matched against the repository description.
            - readme (bool): if True, the search term will be matched against the repository readme.
            - topics (bool): if True, the search term will be matched against the repository topics.
            - owner (str): Specifies the owner of the repositories.
            - repo_name (str): Specifies the name of the repository.
            - user (str): Specifies the user of the repositories.
            - min_size (int): Specifies the minimum size of the repository.
            - max_size (int): Specifies the maximum size of the repository.
            - min_forks (int): Specifies the minimum number of forks of the repository.
            - max_forks (int): Specifies the maximum number of forks of the repository.
            - min_stars (int): Specifies the minimum number of stars of the repository.
            - max_stars (int): Specifies the maximum number of stars of the repository.
            - min_created (datetime.date): Specifies the minimum creation date of the repository.
            - max_created (datetime.date): Specifies the maximum creation date of the repository.
            - language (str): Specifies the language of the repository.
            - topic (str): Specifies the topic of the repository.
            - license (str): Specifies the license of the repository.
    Returns:
        pandas.DataFrame: A dataframe containing information about the fetched repositories.
    """
    qb_obj = GitHubRepoSearchQueryBuilder()
    query_args = get_query_args(**args)
    query_args['value'] = search_term
    qb_obj.init_from_args(query_args)
    query = qb_obj.build()
    print(f"The Full Query is: {query}")

    df = search_github_repos(query,  max_results, token)
    df.to_csv(filename, index=False)
    print(f"Saved to {filename}")
    return df



import os
# from accelerate import Accelerator
from finetuning_datasets import create_datasets
from peft import LoraConfig, get_peft_model#, prepare_model_for_int8_training
from peft import PeftConfig, PeftModel
from finetuning_utils import get_args, SavePeftModelCallback, LoadBestPeftModelCallback
from transformers import AutoModelForCausalLM, AutoTokenizer, Trainer, TrainingArguments, logging, set_seed


def print_trainable_parameters(model):
    """
    Prints the number of trainable parameters in the model.
    """
    trainable_params = 0
    all_param = 0
    for _, param in model.named_parameters():
        all_param += param.numel()
        if param.requires_grad:
            trainable_params += param.numel()
    print(
        f"trainable params: {trainable_params} || all params: {all_param} || trainable%: {100 * trainable_params / all_param}"
    )


def run_training(args, train_data, val_data):
    print("Loading the model")
    if 'ammarnasr/' in args.model_path:    
        config = PeftConfig.from_pretrained(args.model_path)
        model = AutoModelForCausalLM.from_pretrained(
            config.base_model_name_or_path,
            use_auth_token=True,
            use_cache=not args.no_gradient_checkpointing,
            # load_in_8bit=True,
            # device_map={"": Accelerator().process_index},
        )
        # Load the LoRA model
        model = PeftModel.from_pretrained(model, args.model_path, is_trainable=True)
        print_trainable_parameters(model)
    else:

        # disable caching mechanism when using gradient checkpointing
        model = AutoModelForCausalLM.from_pretrained(
            args.model_path,
            use_auth_token=True,
            use_cache=not args.no_gradient_checkpointing,
            # load_in_8bit=True,
            # device_map={"": Accelerator().process_index},
        )
        # model = prepare_model_for_int8_training(model)

        lora_config = LoraConfig(
            r=args.lora_r,
            lora_alpha=args.lora_alpha,
            lora_dropout=args.lora_dropout,
            bias="none",
            task_type="CAUSAL_LM",
            # target_modules = ["c_proj", "c_attn", "q_attn"]
            target_modules = ["qkv_proj"]
        )

        if args.ft_type == "peft":
            print("Using PEFT")
            model = get_peft_model(model, lora_config)

    print_trainable_parameters(model)

    train_data.start_iteration = 0

    print("Starting main loop")
    if args.subset is None:
        args.subset = 'none'
    if 'ammarnasr/' in args.model_path:
        run_name = args.model_path.split('/')[-1]
    else:
        run_name = f"{args.model_path.split('/')[-1]}_{args.dataset_name.split('/')[-1]}_{args.subset.split('/')[-1]}_{args.split}_{args.ft_type}"

    training_args = TrainingArguments(
        output_dir=args.output_dir,
        dataloader_drop_last=True,
        evaluation_strategy="steps",
        max_steps=args.max_steps,
        eval_steps=args.eval_freq,
        save_steps=args.save_freq,
        logging_steps=args.log_freq,
        per_device_train_batch_size=args.batch_size,
        per_device_eval_batch_size=args.batch_size,
        learning_rate=args.learning_rate,
        lr_scheduler_type=args.lr_scheduler_type,
        warmup_steps=args.num_warmup_steps,
        # gradient_accumulation_steps=args.gradient_accumulation_steps,
        # gradient_checkpointing=not args.no_gradient_checkpointing,
        # fp16=not args.no_fp16,
        # bf16=args.bf16,
        weight_decay=args.weight_decay,
        run_name=run_name,
        report_to="wandb",
        push_to_hub=True,
        hub_model_id=f"ammarnasr/{run_name}"
        # ddp_find_unused_parameters=False,
    )

    if args.ft_type == "peft":
        trainer = Trainer(model=model, args=training_args, train_dataset=train_data, eval_dataset=val_data, callbacks=[SavePeftModelCallback, LoadBestPeftModelCallback])
    else:
        trainer = Trainer(model=model, args=training_args, train_dataset=train_data, eval_dataset=val_data)
    print("Training...")
    trainer.train()

    print("Saving last checkpoint of the model")
    model.save_pretrained(os.path.join(args.output_dir, "final_checkpoint/"))


def main(args):
    tokenizer = AutoTokenizer.from_pretrained(args.model_path, use_auth_token=True)
    train_dataset, eval_dataset = create_datasets(tokenizer, args)
    run_training(args, train_dataset, eval_dataset)

def main_from_checkpoint(args):
    tokenizer = AutoTokenizer.from_pretrained(args.model_path, use_auth_token=True)
    train_dataset, eval_dataset = create_datasets(tokenizer, args)
    run_training_from_checkpoint(args, train_dataset, eval_dataset)

def run_training_from_checkpoint(args, train_data, val_data):
    checkpoint_parent_dir = './checkpoints'
    all_checkpoints = os.listdir(checkpoint_parent_dir)
    #remove files that does not start with 'checkpoint-'
    all_checkpoints = [x for x in all_checkpoints if x.startswith('checkpoint-')]
    print(all_checkpoints)
    all_checkpoints_steps = [int(x.split('-')[1]) for x in all_checkpoints]
    index_of_max = all_checkpoints_steps.index(max(all_checkpoints_steps))
    checkpoint_dir = os.path.join(checkpoint_parent_dir, all_checkpoints[index_of_max])
    print("Loading the model from checkpoint: ", checkpoint_dir)
    config = PeftConfig.from_pretrained(checkpoint_dir)
    model = AutoModelForCausalLM.from_pretrained(
        config.base_model_name_or_path,
        use_auth_token=True,
        use_cache=not args.no_gradient_checkpointing,
        # load_in_8bit=True,
        # device_map={"": Accelerator().process_index},
    )
    # Load the LoRA model
    model = PeftModel.from_pretrained(model, checkpoint_dir)

    print_trainable_parameters(model)

    # train_dataset.start_iteration = 0

    print("Starting main loop")
    if args.subset is None:
        args.subset = 'none'
    run_name = f"{args.model_path.split('/')[-1]}_{args.dataset_name.split('/')[-1]}_{args.subset.split('/')[-1]}_{args.split}_{args.ft_type}"


    training_args = TrainingArguments(
        output_dir=args.output_dir,
        dataloader_drop_last=True,
        evaluation_strategy="steps",
        max_steps=args.max_steps,
        eval_steps=args.eval_freq,
        save_steps=args.save_freq,
        logging_steps=args.log_freq,
        per_device_train_batch_size=args.batch_size,
        per_device_eval_batch_size=args.batch_size,
        learning_rate=args.learning_rate,
        lr_scheduler_type=args.lr_scheduler_type,
        warmup_steps=args.num_warmup_steps,
        # gradient_accumulation_steps=args.gradient_accumulation_steps,
        # gradient_checkpointing=not args.no_gradient_checkpointing,
        # fp16=not args.no_fp16,
        # bf16=args.bf16,
        weight_decay=args.weight_decay,
        run_name=run_name,
        report_to="wandb",
        push_to_hub=True,
        hub_model_id=f"ammarnasr/{run_name}"
        # ddp_find_unused_parameters=False,
    )

    trainer = Trainer(model=model, args=training_args, train_dataset=train_data, eval_dataset=val_data, callbacks=[SavePeftModelCallback, LoadBestPeftModelCallback])

    print("Training from checkpoint...", checkpoint_dir, "...")
    trainer.train(checkpoint_dir)


if __name__ == "__main__":
    args = get_args()

    set_seed(args.seed)
    os.makedirs(args.output_dir, exist_ok=True)

    logging.set_verbosity_error()

    main(args)
    # main_from_checkpoint(args)


import pathlib
import setuptools

setuptools.setup(
    name="enigma_ai", 
    version="0.2.1",
    description="Tools for simple and efficient training of LLMs for code generation",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://www.enigma-ai.com/", 
    author="Enigma AI",
    license="MIT",
    project_urls={
        "Source": "https://github.com/ammarnasr/Customizable-Code-Assistant",
        "Documentation": "https://enigma-ai.readthedocs.io/",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",  
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.7",
    install_requires=[
    ],  
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "enigma_ai=enigma_ai.__main__:main"
        ]
    },
)


import base64
import joblib
import requests
from tqdm.auto import tqdm
from ..utils import get_extensions_dict


def get_repo_files(repo_name, branch='master', github_token=None):
    """
    Retrieves a list of file paths in a GitHub repository.

    Args:
        repo_name (str): The full name of the repository (e.g., 'user/repo').
        branch (str, optional): The name of the branch to retrieve files from. Defaults to 'master'.
        github_token (str, optional): The GitHub API token for authentication.

    Returns:
        list: A list of file paths in the repository.
    """
    headers = {}
    headers['Authorization'] = f"token {github_token}"
    url = f"https://api.github.com/repos/{repo_name}/git/trees/{branch}?recursive=1"
    r = requests.get(url, headers=headers)
    if r.status_code == 200:
        res = r.json()
        files = [file["path"] for file in res["tree"]]
        return files
    elif branch == "master":
        files = get_repo_files(repo_name, branch="main", github_token=github_token)
        return files
    else:
        return None
    

def read_github_file(full_name, file_path, github_token=None):
    """
    Reads the content of a file in a GitHub repository.

    Args:
        full_name (str): The full name of the repository (e.g., 'user/repo').
        file_path (str): The path to the file in the repository.
        github_token (str, optional): The GitHub API token for authentication.

    Returns:
        str: The content of the file.
    """
    headers = {}
    if github_token:
        headers['Authorization'] = f"token {github_token}"
        
    url = f'https://api.github.com/repos/{full_name}/contents/{file_path}'
    r = requests.get(url, headers=headers)
    r.raise_for_status()
    data = r.json()
    try:
      file_content = data['content']
      file_content_encoding = data.get('encoding')
      if file_content_encoding == 'base64':
          file_content = base64.b64decode(file_content).decode()
      return file_content
    except:
      pass


def extract_code_from_repos(df, filename, token):
    """
    Extracts code files from the repositories in a dataframe and adds them to a new column.

    Args:
        df (pandas.DataFrame): A dataframe containing repository information.
        filename (str): The filename to use for saving the processed dataframe.
        token (str): The GitHub API token for authentication.

    Returns:
        pandas.DataFrame: The updated dataframe with a new 'code' column containing the extracted code files.
    """
    code_extensions_column_name_dict = get_extensions_dict()
    extracted_filename = filename.replace(".csv", "_with_code.joblib")
    repos_bar_tqdm = tqdm(total=len(df), desc="Extracting Code Files From Repos...", position=0, leave=True)
    all_contents = []
    for _, row in df.iterrows():
        repo_name = row['full_name']
        repo = repo_name
        contents = get_repo_files(repo, github_token=token)
        code_files_dict = {}
        code_files_bar_tqdm = tqdm(total=len(contents), desc=f"Extracting Code Files From {repo}...", position=1, leave=True)
        for code_filepath in contents:
            ext = code_filepath.split('.')[-1]
            if ext not in code_extensions_column_name_dict:
                programming_language = 'Misc'
            else:
                programming_language = code_extensions_column_name_dict[ext]
            raw_content = read_github_file(repo_name, code_filepath, github_token=token)
            if programming_language not in code_files_dict:
                code_files_dict[programming_language] = {}
            code_files_dict[programming_language][code_filepath] = raw_content
            code_files_bar_tqdm.update(1)
        repos_bar_tqdm.update(1)
        all_contents.append(code_files_dict)
    df['code'] = all_contents
    joblib.dump(df, extracted_filename)
    print(f"Finished and saved to {extracted_filename}")
    return df


# finetuneLoRa1B.py
import os
import argparse
import torch
import pandas as pd
import numpy as np
from datasets import load_dataset, DatasetDict
from transformers import AutoModelForCausalLM, AutoTokenizer, Trainer, TrainingArguments
from finetuning_datasets import ConstantLengthDataset
from peft import LoraConfig, get_peft_model
from transformers import TrainerCallback

def main_finetune(main_path, experiment_name, training_data_path):
    

    # Construct the full path for the training data
    training_data_path = os.path.join(main_path, training_data_path)

    # Load the training data
    data = pd.read_csv(training_data_path, low_memory=False)
    # data = data.rename(columns={'code': 'content'})
    shuffled_data = data.sample(frac=1).reset_index(drop=True)
    shuffled_data.to_csv(os.path.join(main_path, 'JS_filesShuffled.csv'), index=False)
    dataset = load_dataset("csv", data_files=os.path.join(main_path, 'JS_filesShuffled.csv'))
    train_test_split = dataset['train'].train_test_split(test_size=0.2)
    test_valid_split = train_test_split['test'].train_test_split(test_size=0.5)
    dataset = DatasetDict({
        'train': train_test_split['train'],
        'validation': test_valid_split['train'],
        'test': test_valid_split['test']
    })

    # Initialize the tokenizer and model
    tokenizer = AutoTokenizer.from_pretrained("Salesforce/codegen2-1B")
    model = AutoModelForCausalLM.from_pretrained("Salesforce/codegen2-1B", trust_remote_code=True, revision="main")
    if tokenizer.pad_token is None:
        tokenizer.pad_token = tokenizer.eos_token
        model.config.pad_token_id = model.config.eos_token_id

    # Tokenize function
    def tokenize_function(examples):
        return tokenizer(examples['content'], truncation=True, padding='max_length', max_length=2048)

    # Apply tokenization to each split
    tokenized_datasets = dataset.map(tokenize_function, batched=True)
    train_ds = tokenized_datasets["train"]
    valid_ds = tokenized_datasets["validation"]
    train_dataset = ConstantLengthDataset(tokenizer, train_ds, infinite=True, seq_length=2048)
    valid_dataset = ConstantLengthDataset(tokenizer, valid_ds, infinite=False, seq_length=2048)

    # Set up LoraConfig and model
    lora_config = LoraConfig(r=64, lora_alpha=128, lora_dropout=0.05, bias='all', task_type='CAUSAL_LM', target_modules=["qkv_proj", "out_proj", "lm_head", "fc_in", "fc_out"])
    model.enable_input_require_grads()
    model = get_peft_model(model, lora_config)

    # Custom callback for early stopping
    class EarlyStoppingCallback(TrainerCallback):
        def __init__(self, early_stopping_patience: int):
            self.early_stopping_patience = early_stopping_patience
            self.best_metric = None
            self.patience_counter = 0

        def on_evaluate(self, args, state, control, metrics=None, **kwargs):
            current_metric = metrics["eval_loss"]
            if self.best_metric is None or current_metric < self.best_metric:
                self.best_metric = current_metric
                self.patience_counter = 0
            else:
                self.patience_counter += 1
                if self.patience_counter >= self.early_stopping_patience:
                    print("Early stopping triggered")
                    control.should_training_stop = True

    # Update TrainingArguments
    training_args = TrainingArguments(
        output_dir=os.path.join(main_path, experiment_name),
        run_name=experiment_name + '-wandb',
        max_steps=10000,
        evaluation_strategy="steps",
        logging_steps=1,
        learning_rate=5e-5,
        warmup_steps=100,
        lr_scheduler_type='cosine',
        gradient_checkpointing=True,
        gradient_accumulation_steps=1,
        per_device_train_batch_size=4,
        per_device_eval_batch_size=4,
        fp16=True,
        load_best_model_at_end=True,
        metric_for_best_model="eval_loss",
        greater_is_better=False,
        save_total_limit=2,
    )

    # Initialize Trainer with early stopping callback
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_dataset,
        eval_dataset=valid_dataset,
        callbacks=[EarlyStoppingCallback(early_stopping_patience=3)]
    )

    # Start training
    trainer.train(resume_from_checkpoint=False)

    # Save the best model
    model_path = os.path.join(main_path, experiment_name)
    if not os.path.exists(model_path):
        try:
            os.makedirs(model_path)
            print(f"Folder '{model_path}' created successfully.")
        except OSError as e:
            print(f"Failed to create folder '{model_path}': {e}")
    else:
        print(f"Folder '{model_path}' already exists.")
    trainer.save_model(model_path)
    print(f"Best model saved to {model_path}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--main_path", required=True, help="Main path where the experiment will be conducted")
    parser.add_argument("--experiment_name", required=True, help="Name of the experiment")
    parser.add_argument("--training_data_path", required=True, help="Path to the training data CSV file")
    args = parser.parse_args()
    main_finetune(args.main_path, args.experiment_name, args.training_data_path)



import os

# Set the GPU (CUDA device) that you want to use
import torch
from dataclasses import dataclass
from datasets import load_dataset
from torch.utils.data import IterableDataset
from torch.utils.data.dataloader import DataLoader
from tqdm import tqdm
from peft import LoraConfig, get_peft_model, PeftConfig, PeftModel
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    Trainer,
    TrainingArguments,
    logging,
    set_seed
)
from finetuning_datasets import ConstantLengthDataset
# model_id = 'Salesforce/codegen-350M-mono'
# tokenizer_id = 'Salesforce/codegen-350M-mono'
lang = 'js'
dataset_id = f'ammarnasr/the-stack-{lang}-clean'
effective_seq_length_train = 2048
effective_seq_length_eval  = 2048
lora_rank = 64
lora_alpha = lora_rank*2
lora_dropout = 0.05
lora_bias = 'all'
lora_task_type = 'CAUSAL_LM'
lora_target_modules = ["qkv_proj", "out_proj", "lm_head", "fc_in", "fc_out"]
dataloader_drop_last = True
max_steps = 10000
eval_steps = 50
save_steps = 100
eval_strategy = 'steps'
logging_steps = 1
learning_rate = 5e-5
warmup_steps = 100
lr_scheduler_type = 'cosine'
gradient_checkpointing = True
gradient_accumulation_steps = 1
per_device_train_batch_size  = 4
per_device_eval_batch_size = 4
fp16 = True

exp_name = f"codegen-{lang}-LoRa-v7-run-1"

tokenizer = AutoTokenizer.from_pretrained("Salesforce/codegen2-1B")
model = AutoModelForCausalLM.from_pretrained("Salesforce/codegen2-1B", trust_remote_code=True, revision="main")

###
import pandas as pd
import numpy as np

# Read the CSV file
data = pd.read_csv("/PATH/JS_files.csv")

# Rename the 'code' column to 'content'
data = data.rename(columns={'code': 'content'})

# Fix the random seed for reproducibility
random_seed = 42
np.random.seed(random_seed)

# Shuffle the dataset
shuffled_data = data.sample(frac=1).reset_index(drop=True)

# Save the shuffled data to a new CSV file
shuffled_data.to_csv('/PATH/JS_filesSuffled.csv', index=False)


from transformers import AutoTokenizer
from datasets import DatasetDict, load_dataset

# Load the dataset from a CSV file
dataset = load_dataset("csv", data_files="/PATH/JS_filesSuffled.csv")

# Split the dataset into train, validation, and test sets
train_test_split = dataset['train'].train_test_split(test_size=0.2)
test_valid_split = train_test_split['test'].train_test_split(test_size=0.5)
dataset = DatasetDict({
    'train': train_test_split['train'],
    'validation': test_valid_split['train'],
    'test': test_valid_split['test']
})

effective_seq_length_train = 2048
effective_seq_length_eval = 2048

# The CodeGen tokenizer always during exceuction lacks the padding token if the tokenizer has a padding token and add one if not
if tokenizer.pad_token is None:
    tokenizer.pad_token = tokenizer.eos_token
    model.config.pad_token_id = model.config.eos_token_id
# Tokenize function
def tokenize_function(examples):
    return tokenizer(examples['content'], truncation=True, padding='max_length', max_length=effective_seq_length_train)

# Apply tokenization to each split
tokenized_datasets = dataset.map(tokenize_function, batched=True)

# Check the tokenized dataset
print(tokenized_datasets['train'][0])

#########
train_ds = tokenized_datasets["train"]
valid_ds = tokenized_datasets["validation"]
print(train_ds[0])

train_dataset = ConstantLengthDataset(tokenizer, train_ds, infinite=True, seq_length=effective_seq_length_train)
valid_dataset = ConstantLengthDataset(tokenizer, valid_ds, infinite=False, seq_length=effective_seq_length_eval)
###

#===================================
lora_config = LoraConfig(r = lora_rank, lora_alpha = lora_alpha, lora_dropout = lora_dropout, bias = lora_bias, task_type = lora_task_type, target_modules = lora_target_modules)
model.enable_input_require_grads()
model = get_peft_model(model, lora_config)
model.print_trainable_parameters()
#===================================
from transformers import AutoTokenizer, AutoModelForCausalLM
# tokenizer = AutoTokenizer.from_pretrained("Salesforce/codegen2-1B")
# model = AutoModelForCausalLM.from_pretrained("Salesforce/codegen2-1B", trust_remote_code=True, revision="main")

from transformers import TrainerCallback

# Custom callback for early stopping
class EarlyStoppingCallback(TrainerCallback):
    def __init__(self, early_stopping_patience: int):
        self.early_stopping_patience = early_stopping_patience
        self.best_metric = None
        self.patience_counter = 0

    def on_evaluate(self, args, state, control, metrics=None, **kwargs):
        current_metric = metrics["eval_loss"]  # Change this to your evaluation metric
        if self.best_metric is None or current_metric < self.best_metric:
            self.best_metric = current_metric
            self.patience_counter = 0
        else:
            self.patience_counter += 1
            if self.patience_counter >= self.early_stopping_patience:
                print("Early stopping triggered")
                control.should_training_stop = True
exp_name = "CodeGen2_1BEarlyStopping"
# Update TrainingArguments
training_args = TrainingArguments(
    output_dir=exp_name,
    run_name=exp_name+'-wandb',
    dataloader_drop_last=dataloader_drop_last,
    max_steps=max_steps,
    eval_steps=eval_steps,
    save_steps=save_steps,
    evaluation_strategy=eval_strategy,
    logging_steps=logging_steps,
    learning_rate=learning_rate,
    warmup_steps=warmup_steps,
    lr_scheduler_type=lr_scheduler_type,
    gradient_checkpointing=gradient_checkpointing,
    gradient_accumulation_steps=gradient_accumulation_steps,
    per_device_train_batch_size=per_device_train_batch_size,
    per_device_eval_batch_size=per_device_eval_batch_size,
    fp16=fp16,
    load_best_model_at_end=True,  # Load best model at the end of training
    metric_for_best_model="eval_loss",  # Change this to your evaluation metric
    greater_is_better=False,  # Set to True if higher metric is better
    save_total_limit=2,  # Save only 2 checkpoints
)

# Initialize Trainer with early stopping callback
trainer = Trainer(
    model=model,
    args=training_args,
    train_dataset=train_dataset,
    eval_dataset=valid_dataset,
    callbacks=[EarlyStoppingCallback(early_stopping_patience=3)]  # Set your patience
)

# Start training
trainer.train(resume_from_checkpoint=False)
# After training, save the best model
model_path = "/PATH/CodeGen2_1BEarlyStoppingJS/best_model"
import os

model_path = "/PATH/CodeGen2_1BEarlyStoppingJS/best_model"

# Check if the folder exists
if not os.path.exists(model_path):
    # Create the folder if it doesn't exist
    try:
        os.makedirs(model_path)
        print(f"Folder '{model_path}' created successfully.")
    except OSError as e:
        print(f"Failed to create folder '{model_path}': {e}")
else:
    print(f"Folder '{model_path}' already exists.")

trainer.save_model(model_path)

print(f"Best model saved to {model_path}")